#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif

#include "GLSL_helper.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
#include "Song.hpp"
#include <cmath>
#include "SOIL.h"
#include "MStackHelp.h"
#include <string>
#include <sstream>
#include <cstdlib>
#define MOVE_UNIT_MS 1000


enum Visualization {COMMENT_LENGTH, NUM_WORDS};

int ShadeProg[2];
static float  g_width, g_height;

//Handles to the shader data
GLint h_uTexUnit;
GLint h_aPosition;
GLint h_aTexCoord;
GLint h_uModelMatrix;
GLint h_uViewMatrix;
GLint h_uProjMatrix;
GLint h_uWaveRender;




GLuint *textures;
GLuint *avatar_textures;
GLuint g_Placeholder;

GLint h_aPositionNoTex;
GLint h_uModelMatrixNoTex;
GLint h_uViewMatrixNoTex;
GLint h_uProjMatrixNoTex;
GLint h_uColorNoTex;

// Camera Stuff
int g_startMouseX, g_startMouseY;
float g_LookAtX, g_LookAtY, g_LookAtZ;
float g_PrevAlpha, g_PrevBeta;
float g_Alpha, g_Beta;
float g_MoveSpeed = 0.2f;

glm::vec3 g_EyePos = glm::vec3(-5.0f, 0.5f, 0);
glm::vec3 g_LookAtPos = glm::vec3(-5.0f, 0.5f, -1.0f);
glm::vec3 g_UpVec = glm::vec3(0, 1.0, 0);
glm::vec3 camW = glm::vec3(0, 0, -1.0f);
glm::vec3 camU = glm::vec3(-1.0f, 0, 0);

int g_NearestComment = 0;


// Waveform buffer stuff
GLuint WaveformBuffObj, WIndxBuffObj, WaveformTexBuffObj;
int g_WiboLen;


// Cube buffer stuff
GLuint VisualizeBuffObj, VIndxBuffObj;
int g_ViboLen;

float cameraMoveRight = 0.0;
glm::vec3 cameraMoveVector = glm::vec3(1.0, 0, 0);
glm::vec3 moveFactor = glm::vec3(0, 0, 0);

float waveFormWidth = 10.f;
float cameraMoveUnit = .01f;
float songMoveUnit;
float songTimeMS = 0;

int g_MaxCommentLength;
int g_MaxCommentWords;
float g_VisualizationScaleFactor = 1.0f;

song_t * song;

//declare a matrix stack
RenderingHelper ModelTrans;


float g_RotateAvatars = 0;

int currentComments[4];
bool pause = false;
Visualization g_DataToRender = COMMENT_LENGTH;
bool g_RenderBoth = false;
bool reverse = false;

int dataToUse = 1;


void setupCameraMovement()
{
  float msPerWavelength = (float)(song->lengthInMS) / waveFormWidth;
  // How many ms in the song do we move per camera move?
  songMoveUnit = msPerWavelength * cameraMoveUnit;
  std::cout << "songMoveUnit: " << songMoveUnit << "\n";
}

void updateMoveFactor()
{
   moveFactor = cameraMoveVector * cameraMoveRight;
}

/* projection matrix */
void SetProjectionMatrix(int ShaderNumber) {
  glm::mat4 Projection = glm::perspective(80.0f, (float)g_width/g_height, 0.1f, 100.f);
  if(ShaderNumber == 0) {
      safe_glUniformMatrix4fv(h_uProjMatrix, glm::value_ptr(Projection));
  } else {
      safe_glUniformMatrix4fv(h_uProjMatrixNoTex, glm::value_ptr(Projection));
  }

}


void SetView(int ShaderNumber) {
   glm::vec3 g_TempEyePos = g_EyePos + moveFactor;
   glm::vec3 g_TempLookAtPos = g_LookAtPos + moveFactor;

   glm::mat4 view = glm::lookAt(g_TempEyePos, g_TempLookAtPos,g_UpVec);
   if(ShaderNumber == 0) {
     safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(view));
   } else {
       safe_glUniformMatrix4fv(h_uViewMatrixNoTex, glm::value_ptr(view));
   }
}

/* set the model transform to the identity */
void SetModelWaveform() {
  glm::mat4 tmp = glm::mat4(1.0f);
  glm::mat4 Scale = glm::scale(tmp, glm::vec3(waveFormWidth, 1.0f, 1.0f));
  glm::mat4 Trans = glm::translate(Scale, glm::vec3(0, -1.0f, -1.85f));
  safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(Trans));
}

void SetModelArtwork() {
   glm::mat4 tmp = glm::mat4(1.0f);
   glm::mat4 Scale = glm::scale(tmp, glm::vec3(20.0f, 20.0f, 1.0f));
   glm::mat4 Trans = glm::translate(Scale, glm::vec3(0, -0.5f, -5.0f));
   safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(Trans));
}

void setModelVisualize()
{
  glm:: mat4 tmp = glm::mat4(1.0f);
  glm::mat4 Trans = glm::translate(tmp, glm::vec3(-4.0f, 0, -3.0f) + moveFactor);
  safe_glUniformMatrix4fv(h_uModelMatrixNoTex, glm::value_ptr(Trans));
}

void setModelUserAvatar()
{
  glm:: mat4 tmp = glm::mat4(1.0f);
  glm::mat4 Trans = glm::translate(tmp, glm::vec3(-5.0f, 0, -3.0f) + moveFactor);
  safe_glUniformMatrix4fv(h_uModelMatrixNoTex, glm::value_ptr(Trans));
}

/* model transforms - do not change - sets matrix in shader */
void SetModel() {
  safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
}

static void initVisualizationGeom()
{
    float VisualizePos[] = {
      // First Triangle
      -0.5, 0.0, 0.0,  // BLeft 
      0.5, 0.0, 0.0,   // BRight
      -0.5, 1.0, 0.0,  // TLeft
      // Second Triangle
      -0.5, 1.0, 0.0,  // TLeft
      0.5, 1.0, 0.0,   // TRight
      0.5, 0.0, 0.0    // BRight

   };

  unsigned short idx[] = {0, 1, 2, 3, 4, 5};

  g_ViboLen = 6;
  glGenBuffers(1, &VisualizeBuffObj);
  glBindBuffer(GL_ARRAY_BUFFER, VisualizeBuffObj);
  glBufferData(GL_ARRAY_BUFFER, sizeof(VisualizePos), VisualizePos, GL_STATIC_DRAW);

  glGenBuffers(1, &VIndxBuffObj);
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, VIndxBuffObj);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);


}



static void initWaveformBillboard() {

  float WaveformPos[] = {
    // First Triangle
    -0.5, 0.0, 0.0,  // BLeft 
    0.5, 0.0, 0.0,   // BRight
    -0.5, 1.0, 0.0,  // TLeft
    // Second Triangle
    -0.5, 1.0, 0.0,  // TLeft
    0.5, 1.0, 0.0,   // TRight
    0.5, 0.0, 0.0    // BRight

 };

 static GLfloat WaveformTex[] = {
  0, 0,
  1.0, 0,
  0, 1.0,

  0, 1.0,
  1.0, 1.0,
  1.0, 0
}; 

unsigned short idx[] = {0, 1, 2, 3, 4, 5};

g_WiboLen = 6;
glGenBuffers(1, &WaveformBuffObj);
glBindBuffer(GL_ARRAY_BUFFER, WaveformBuffObj);
glBufferData(GL_ARRAY_BUFFER, sizeof(WaveformPos), WaveformPos, GL_STATIC_DRAW);

glGenBuffers(1, &WIndxBuffObj);
glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, WIndxBuffObj);
glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);

glGenBuffers(1, &WaveformTexBuffObj);
glBindBuffer(GL_ARRAY_BUFFER, WaveformTexBuffObj);
glBufferData(GL_ARRAY_BUFFER, sizeof(WaveformTex), WaveformTex, GL_STATIC_DRAW);

}

void InitGeom() {
  initWaveformBillboard();
  initVisualizationGeom();
}

/*function to help load the shaders (both vertex and fragment */
int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName, int ShaderNumber) {
  GLuint VS; //handles to shader object
        GLuint FS; //handles to frag shader object
        GLint vCompiled, fCompiled, linked; //status of shader

        VS = glCreateShader(GL_VERTEX_SHADER);
        FS = glCreateShader(GL_FRAGMENT_SHADER);

        //load the source
        glShaderSource(VS, 1, &vShaderName, NULL);
        glShaderSource(FS, 1, &fShaderName, NULL);
        //compile shader and print log
        glCompileShader(VS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
        printShaderInfoLog(VS);

        //compile shader and print log
        glCompileShader(FS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
        printShaderInfoLog(FS);

        if (!vCompiled || !fCompiled) {
          printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
          return 0;
        }

        //create a program object and attach the compiled shader
        ShadeProg[ShaderNumber] = glCreateProgram();
        glAttachShader(ShadeProg[ShaderNumber], VS);
        glAttachShader(ShadeProg[ShaderNumber], FS);

        glLinkProgram(ShadeProg[ShaderNumber]);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetProgramiv(ShadeProg[ShaderNumber], GL_LINK_STATUS, &linked);
        printProgramInfoLog(ShadeProg[ShaderNumber]);

        glUseProgram(ShadeProg[ShaderNumber]);
        if(ShaderNumber == 0) {
        /* get handles to attribute data */
          h_aPosition = safe_glGetAttribLocation(ShadeProg[ShaderNumber], "aPosition");
          h_aTexCoord = safe_glGetAttribLocation(ShadeProg[ShaderNumber],  "aTexCoord");
          h_uTexUnit = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uTexUnit");
          h_uProjMatrix = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uProjMatrix");
          h_uViewMatrix = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uViewMatrix");
          h_uModelMatrix = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uModelMatrix");
          h_uWaveRender = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uWaveRender");
        }
        if(ShaderNumber == 1) {
          /* get handles to attribute data */
          h_aPositionNoTex = safe_glGetAttribLocation(ShadeProg[ShaderNumber], "aPosition");
          h_uProjMatrixNoTex = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uProjMatrix");
          h_uViewMatrixNoTex = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uViewMatrix");
          h_uModelMatrixNoTex = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uModelMatrix");
          h_uColorNoTex = safe_glGetUniformLocation(ShadeProg[ShaderNumber], "uColor");
        }

        printf("sucessfully installed shader %d\n", ShadeProg[ShaderNumber]);
        return 1;

}

void setShader(int ShaderNumber)
{
  glUseProgram(ShadeProg[ShaderNumber]);
}

/* Some OpenGL initialization */
void Initialize ()                  // Any GL Init Code
{
    // Start Of User Initialization
    glClearColor (1.0f, 1.0f, 1.0f, 1.0f);
    // Black Background
    glClearDepth (1.0f);    // Depth Buffer Setup
    glDepthFunc (GL_LEQUAL);    // The Type Of Depth Testing
    glEnable (GL_DEPTH_TEST);// Enable Depth Testing

    /* texture specific settings */
    glEnable(GL_TEXTURE_2D);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);


    textures = new GLuint[3];
    glGenTextures(3, textures);
   ModelTrans.useModelViewMatrix();
 }

/* Reshape */
 void ReshapeGL (int width, int height)
 {
    g_width = (float)width;
    g_height = (float)height;
    glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));

 }

 void DrawWaveform()
 {
    //Set up matrix transforms
    SetProjectionMatrix(0);
    SetView(0);
    SetModelWaveform();

    //set up the texture unit
    glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, textures[0]);

    safe_glUniform1i(h_uTexUnit, 0);
    safe_glUniform1i(h_uWaveRender, 1);

    safe_glEnableVertexAttribArray(h_aPosition);
    glBindBuffer(GL_ARRAY_BUFFER, WaveformBuffObj);
    safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

    safe_glEnableVertexAttribArray(h_aTexCoord);
    glBindBuffer(GL_ARRAY_BUFFER, WaveformTexBuffObj);
    safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 

    // bind ibo
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, WIndxBuffObj);
    glDrawElements(GL_TRIANGLES, g_WiboLen, GL_UNSIGNED_SHORT, 0);

    safe_glDisableVertexAttribArray(h_aPosition);
    safe_glDisableVertexAttribArray(h_aTexCoord);
 }

 void loadUserImage(int index)
 {
    if(index > song->numberOfComments - 1) {
       avatar_textures[index] = 0;
       return;
    }
    std::ostringstream os;
    os << "./SongData" << dataToUse << "/UserAvatars/" << index << ".jpg";
    const std::string avatar_filename = os.str();
    const char *filename = avatar_filename.c_str();

    avatar_textures[index] = SOIL_load_OGL_texture
    (
     filename,
     SOIL_LOAD_AUTO,
     SOIL_CREATE_NEW_ID,
     SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
     );
    /*
    if( 0 == avatar_textures[index] )
    {
      printf( "SOIL loading error: '%s', %s \n", SOIL_last_result(),  filename);
    }
    */
 }

 void setupAvatar(int add)
 {
   bool usePlaceholder = false;
   if(!(song->comments[g_NearestComment + add].avatarLoaded)) {
      loadUserImage(g_NearestComment + add);
      if(0 == avatar_textures[g_NearestComment + add]) {
         usePlaceholder = true;
      } else {
         song->comments[g_NearestComment + add].avatarLoaded = true;
      }
   }
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);
   if(!usePlaceholder) {
      glBindTexture(GL_TEXTURE_2D, avatar_textures[g_NearestComment + add]);
   } else {
      glBindTexture(GL_TEXTURE_2D, g_Placeholder);
   }

    safe_glUniform1i(h_uTexUnit, 0);
    safe_glUniform1i(h_uWaveRender, 0);

    safe_glEnableVertexAttribArray(h_aPosition);
    glBindBuffer(GL_ARRAY_BUFFER, WaveformBuffObj);
    safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

    safe_glEnableVertexAttribArray(h_aTexCoord);
    glBindBuffer(GL_ARRAY_BUFFER, WaveformTexBuffObj);
    safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 

    // bind ibo
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, WIndxBuffObj);
    glDrawElements(GL_TRIANGLES, g_WiboLen, GL_UNSIGNED_SHORT, 0);

    safe_glDisableVertexAttribArray(h_aPosition);
    safe_glDisableVertexAttribArray(h_aTexCoord);
 }

void setupCommentLength()
{
   safe_glEnableVertexAttribArray(h_aPositionNoTex);
   glBindBuffer(GL_ARRAY_BUFFER, VisualizeBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

    // bind ibo
   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, VIndxBuffObj);
   glDrawElements(GL_TRIANGLES, g_ViboLen, GL_UNSIGNED_SHORT, 0);

   safe_glDisableVertexAttribArray(h_aPositionNoTex);
}


void Draw4Avatars()
{
   SetProjectionMatrix(1);
   SetView(1);
   glEnable(GL_TEXTURE_2D);
   glActiveTexture(GL_TEXTURE0);


   ModelTrans.loadIdentity();
   ModelTrans.pushMatrix();
      ModelTrans.translate(glm::vec3(-5.0f, 0, -3.0f) + moveFactor);
      ModelTrans.rotate(g_RotateAvatars, glm::vec3(0, 1.0, 0));
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(0, 0, .5f));
         SetModel();
         setupAvatar(0);
      ModelTrans.popMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(.5f, 0, 0));
         ModelTrans.rotate(90.0, glm::vec3(0, 1.0, 0));
         SetModel();
         setupAvatar(1);
      ModelTrans.popMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(0, 0, -.5f));
         ModelTrans.rotate(180.0, glm::vec3(0, 1.0, 0));
         SetModel();
         setupAvatar(2);
      ModelTrans.popMatrix();
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(-.5f, 0, 0));
         ModelTrans.rotate(-90.0, glm::vec3(0, 1.0, 0));        
         SetModel();
         setupAvatar(3);
      ModelTrans.popMatrix();
   ModelTrans.popMatrix();
}

void calculateMaxValues()
{
   g_MaxCommentLength = song->comments[0].commentLength;
   g_MaxCommentWords = song->comments[0].numberOfWords;
   for(int i = 1; i < song->numberOfComments; i++) {
      if(song->comments[i].commentLength > g_MaxCommentLength) {
         g_MaxCommentLength = song->comments[i].commentLength;
      }
      if(song->comments[i].numberOfWords > g_MaxCommentWords) {
         g_MaxCommentWords = song->comments[i].numberOfWords;
      }      
   }
}

float yScaleVisualization(Visualization v, int commentIndex) {
   if(v == COMMENT_LENGTH) {
      return (float)(song->comments[commentIndex].commentLength) / (float)g_MaxCommentLength * g_VisualizationScaleFactor;
   } else if (v == NUM_WORDS) {
      return (float)(song->comments[commentIndex].numberOfWords) / (float)g_MaxCommentWords * g_VisualizationScaleFactor;
   }
   return 1.0f;
}

void Draw4CommentLengths(Visualization v)
{
   SetProjectionMatrix(1);
   SetView(1);
   if(v == COMMENT_LENGTH) {
      glUniform3f(h_uColorNoTex, 0, 1.0, 0);
   } else {
      glUniform3f(h_uColorNoTex, 1.0, 0, 0);
   }

   float xScale = g_RenderBoth ? 0.5f : 1.0f;
   float extraShift = g_RenderBoth && v == NUM_WORDS ? 0.5f : 0;

   ModelTrans.loadIdentity();
   ModelTrans.pushMatrix();
      ModelTrans.translate(glm::vec3(-5.0f, 0, -3.0f) + moveFactor);
      ModelTrans.rotate(g_RotateAvatars, glm::vec3(0, 1.0, 0));
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(.5 + extraShift, 0, .5f));
         //ModelTrans.scale(1.0, 2.0, 1.0);
         ModelTrans.scale(xScale, yScaleVisualization(v,currentComments[0]),1.0);
         SetModel();
         setupCommentLength();
      ModelTrans.popMatrix();
      
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(0.5f, 0, -.5f - extraShift));
         ModelTrans.rotate(90.0, glm::vec3(0, 1.0, 0));
         ModelTrans.scale(xScale, yScaleVisualization(v,currentComments[1]),1.0);
         SetModel();
         setupCommentLength();
      ModelTrans.popMatrix();
      
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(-.5 - extraShift, 0, -0.5f));
         ModelTrans.rotate(180.0, glm::vec3(0, 1.0, 0));
         ModelTrans.scale(xScale, yScaleVisualization(v,currentComments[2]),1.0);
         SetModel();
         setupCommentLength();
      ModelTrans.popMatrix();
      
      
      ModelTrans.pushMatrix();
         ModelTrans.translate(glm::vec3(-.5f, 0, .5f + extraShift));
         ModelTrans.rotate(-90.0, glm::vec3(0, 1.0, 0));  
         ModelTrans.scale(xScale, yScaleVisualization(v,currentComments[3]),1.0);     
         SetModel();
         setupCommentLength();
      ModelTrans.popMatrix();
      
      
   ModelTrans.popMatrix();
}


void DrawUserAvatar(int add)
{

 //Set up matrix transforms
 SetProjectionMatrix(0);
 SetView(0);
 setModelUserAvatar();

 //set up the texture unit
 glEnable(GL_TEXTURE_2D);
 glActiveTexture(GL_TEXTURE0);
 glBindTexture(GL_TEXTURE_2D, avatar_textures[g_NearestComment + add]);

 safe_glUniform1i(h_uTexUnit, 0);
 safe_glUniform1i(h_uWaveRender, 0);

 safe_glEnableVertexAttribArray(h_aPosition);
 glBindBuffer(GL_ARRAY_BUFFER, WaveformBuffObj);
 safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

 safe_glEnableVertexAttribArray(h_aTexCoord);
 glBindBuffer(GL_ARRAY_BUFFER, WaveformTexBuffObj);
 safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 

 // bind ibo
 glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, WIndxBuffObj);
 glDrawElements(GL_TRIANGLES, g_WiboLen, GL_UNSIGNED_SHORT, 0);

 safe_glDisableVertexAttribArray(h_aPosition);
 safe_glDisableVertexAttribArray(h_aTexCoord);
}

void DrawVisualize()
{
  SetProjectionMatrix(1);
  SetView(1);
  setModelVisualize();

  safe_glEnableVertexAttribArray(h_aPositionNoTex);
  glBindBuffer(GL_ARRAY_BUFFER, VisualizeBuffObj);
  safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

   // bind ibo
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, VIndxBuffObj);
  glDrawElements(GL_TRIANGLES, g_ViboLen, GL_UNSIGNED_SHORT, 0);

  safe_glDisableVertexAttribArray(h_aPositionNoTex);

}


void DrawArtwork()
{

   //Set up matrix transforms
  SetProjectionMatrix(0);
  SetView(0);
  SetModelArtwork();


    //set up the texture unit
  glEnable(GL_TEXTURE_2D);
  glActiveTexture(GL_TEXTURE0);
  glBindTexture(GL_TEXTURE_2D, textures[1]);

  safe_glUniform1i(h_uTexUnit, 0);
  safe_glUniform1i(h_uWaveRender, 0);

  safe_glEnableVertexAttribArray(h_aPosition);
  glBindBuffer(GL_ARRAY_BUFFER, WaveformBuffObj);
  safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

  safe_glEnableVertexAttribArray(h_aTexCoord);
  glBindBuffer(GL_ARRAY_BUFFER, WaveformTexBuffObj);
  safe_glVertexAttribPointer(h_aTexCoord, 2, GL_FLOAT, GL_FALSE, 0, 0); 

    // bind ibo
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, WIndxBuffObj);
  glDrawElements(GL_TRIANGLES, g_WiboLen, GL_UNSIGNED_SHORT, 0);

  safe_glDisableVertexAttribArray(h_aPosition);
  safe_glDisableVertexAttribArray(h_aTexCoord);
}


 void Draw()
 {
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    setShader(0);

    glEnable(GL_TEXTURE_2D);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    DrawWaveform();
    DrawArtwork();
    Draw4Avatars();
    glDisable(GL_TEXTURE_2D);

    
   setShader(1);
   if(!g_RenderBoth) {
      Draw4CommentLengths(g_DataToRender);
   } else {
      Draw4CommentLengths(COMMENT_LENGTH);
      Draw4CommentLengths(NUM_WORDS);
   }
    
    //Disable the shader
    glUseProgram(0);
    glutSwapBuffers();
 }

glm::vec3 cameraWVector()
{
   glm::vec3 gaze = g_LookAtPos - g_EyePos;
   return glm::normalize(-1.0f * gaze);
}

glm::vec3 cameraUVector()
{
   return glm::normalize(glm::cross(g_UpVec,cameraWVector()));
}

void strafe(bool left)
{
   float dir = left ? 1.0f : -1.0f;
   glm::vec3 tmp = camU * g_MoveSpeed * dir;
   g_EyePos += tmp;
   g_LookAtPos += tmp;
}

void forwardAndBack(bool forward)
{
   float dir = forward ? 1.0f : -1.0f;
   glm::vec3 tmp = camW * g_MoveSpeed * dir;
   g_EyePos += tmp;
   g_LookAtPos += tmp;
}

void updateCurrentComments()
{
   for(int i = 0; i < 4; i++) {
      currentComments[i] = g_NearestComment + i;
   }
}

void updateNearestComment()
{
  if(g_NearestComment + 4 == song->numberOfComments - 1) {
    std::cout << "At last user.\n";
    return;
  }
  if(std::abs(songTimeMS - song->comments[g_NearestComment].commentTime) >=
    std::abs(songTimeMS - song->comments[g_NearestComment + 1].commentTime)) {
   glDeleteTextures(1, &(avatar_textures[g_NearestComment]));
   g_NearestComment += 1;
   updateCurrentComments();
   return;
   }
}    


void moveAlongWaveform(int value)
{
   if(pause) {
      return;
   }
   float sign = reverse ? -1.0f : 1.0f;
   songTimeMS += songMoveUnit * sign;
   cameraMoveRight += cameraMoveUnit * sign;
   updateMoveFactor();
   updateNearestComment();
   if(songTimeMS < (float)(song->lengthInMS)) {
      if((reverse && !(songTimeMS < 0)) || !reverse)
         glutTimerFunc(songMoveUnit, moveAlongWaveform, 1);
   }
   glutPostRedisplay();
}

void updateAvatarRotation(int value)
{
   if(pause) return;
   float sign = reverse ? -1.0f : 1.0f;
   g_RotateAvatars += 2 * sign;
   if(g_RotateAvatars >= 360.0) {
      g_RotateAvatars = 0;
   }
   if(g_RotateAvatars < 0) {
      g_RotateAvatars = 360.0;
   }
   if(songTimeMS < (float)(song->lengthInMS)) {
      if((reverse && ! (songTimeMS < 0))|| !reverse)
         glutTimerFunc(30, updateAvatarRotation, 0);
   }
   glutPostRedisplay();
}

void swapDataToRender()
{
   if(g_DataToRender == COMMENT_LENGTH) g_DataToRender = NUM_WORDS;
   else g_DataToRender = COMMENT_LENGTH;
}

//the keyboard callback
void keyboard(unsigned char key, int x, int y ){
  switch( key ) {
    case 'w':
      forwardAndBack(true);
      break;
    case 's':
      forwardAndBack(false);
    break;
    case 'a':
      strafe(true);
      break;
    case 'd':
      strafe(false);
      break;
    case 'p':
      pause = !pause;
      if(!pause) {
         glutTimerFunc(songMoveUnit, moveAlongWaveform, 1);
         glutTimerFunc(30, updateAvatarRotation, 0);
      }
      break;
    case '1':
      swapDataToRender();
      break;
    case '2':
      g_RenderBoth = !g_RenderBoth;
      return;
    case 'r':
      reverse = !reverse;
      std::cout << "Songtime " << songTimeMS << "\n"; 
      break;
    case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
      break;
  }
  glutPostRedisplay();
}


void loadArtwork()
{
   std::ostringstream os;
   os << "./SongData" << dataToUse << "/artwork.jpg";
   const std::string filename = os.str();

  textures[1] = SOIL_load_OGL_texture
  (
    filename.c_str(),
    SOIL_LOAD_AUTO,
    SOIL_CREATE_NEW_ID,
    SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
    );
}

void loadWaveform()
{
   std::ostringstream os;
   os << "./SongData" << dataToUse << "/waveform.png";
   const std::string filename = os.str();
   std::cout << "Waveform path: " << filename << "\n";
  textures[0] = SOIL_load_OGL_texture
  (
    filename.c_str(),
    SOIL_LOAD_AUTO,
    SOIL_CREATE_NEW_ID,
    SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
    );
}

void loadUserImages()
{
  for(int i = 0; i < song->numberOfComments; i++) {
    std::ostringstream os;
    os << "./SongData/UserAvatars/" << i << ".jpg";
    const std::string avatar_filename = os.str();
    const char *filename = avatar_filename.c_str();

    avatar_textures[i] = SOIL_load_OGL_texture
    (
      filename,
      SOIL_LOAD_AUTO,
      SOIL_CREATE_NEW_ID,
      SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
      );
  }
}

void loadPlaceholder()
{
     g_Placeholder = SOIL_load_OGL_texture
     (
       "./SongData/avatar_placeholder.jpg",
       SOIL_LOAD_AUTO,
       SOIL_CREATE_NEW_ID,
       SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
       );
}

 int main(int argc, char **argv) {
  //initialize the window
  glutInit(&argc, argv);
  if(argc < 2) {
   dataToUse = 1;
  } else {
     dataToUse = atoi(argv[1]);
  }
  std::ostringstream os;
  os << "./SongData" << dataToUse << "/song_comments.txt";
  std::string path = os.str();
  const char *pathname = path.c_str();
  song = parseSongDataFile(pathname);
  calculateMaxValues();
  std::cout << song->numberOfComments << "\n";
  std::cout << song->lengthInMS << "\n";
  songTimeMS = 0;
  avatar_textures = new GLuint[song->numberOfComments];
  setupCameraMovement();
  glutInitWindowPosition( 20, 20 );
  glutInitWindowSize(600, 600);
  glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutCreateWindow(song->title.c_str());

  //set up the opengl call backs
  glutDisplayFunc(Draw);
  glutReshapeFunc(ReshapeGL);
  glutKeyboardFunc( keyboard );

  glutTimerFunc(songMoveUnit, moveAlongWaveform, 0);
  glutTimerFunc(30, updateAvatarRotation, 0);

  Initialize();
  loadWaveform();
  loadArtwork();
  loadPlaceholder();
  //loadUserImages();

  //test the openGL version
  getGLversion();
  //install the shader

  if (!InstallShader(textFileRead((char *)"tex_vert.glsl"), textFileRead((char *)"tex_frag.glsl"), 0)) {
        printf("Error installing shader!\n");
        return 0;
  }
  if (!InstallShader(textFileRead((char *)"test_vert.glsl"), textFileRead((char *)"test_frag.glsl"), 1)) {
        printf("Error installing shader!\n");
        return 0;
  }
  InitGeom();
  
  glutMainLoop();


}