#!/usr/bin/python
import requests as req
import soundcloud
import sys
from nltk import word_tokenize


class SongDataParser:
   _data_path = "./SongData3/"

   def __init__(self):
      self._login()

   def _login(self):
      self.client = soundcloud.Client(client_id='',
         client_secret='',
         username='',
         password='')

   def _scrape_artwork(self, song):
      artwork_url = song.artwork_url
      artwork_url = artwork_url.rsplit("?")[0]
      artwork_url = artwork_url.replace('large.jpg', 'original.jpg')
      print artwork_url
      with open(self._data_path + "artwork", 'wb') as f:
         f.write(req.get(artwork_url).content)

   def _scrape_waveform(self, song):
      waveform_url = song.waveform_url
      wave_file_name = waveform_url.rsplit("/", 1)[1]
      with open(self._data_path + "waveform", 'wb') as f:
         f.write(req.get(waveform_url).content)

   def _scrape_comments(self, song):
      index = 0
      with open("./SongData3/song_comments.txt", 'w') as f:
         comments = retrieve_song_comments(song.id, self.client)
         f.write(str(song.user[u'username']) + " - " + str(song.title) + "\n")
         f.write(str(song.duration) + "\n")
         f.write(str(len(comments)) + "\n")
         for comment in sorted(comments, key=lambda comment: comment.timestamp):
            if comment.timestamp is None:

               continue
            f.write(str(comment.timestamp) + " ")
            f.write(str(len(comment.body)) + " ")
            f.write(str(len(word_tokenize(comment.body))) + "\n")
            with open("./SongData3/UserAvatars/" + str(index) + ".jpg", "w") as avatar:
               avatar.write(req.get(comment.user[u"avatar_url"]).content)
            index += 1


   def scrape_song(self, url):
      song = self.client.get('/resolve', url=url)
      self._scrape_waveform(song)
      self._scrape_artwork(song)
      self._scrape_comments(song)



def retrieve_song_comments(songid, client):
    offset = 0
    limit = 200
    all_comments_scraped = False
    all_comments = []
    while not all_comments_scraped:
        next_round_of_comments = client.get('/tracks/%d/comments' % songid, limit=limit, offset=offset)
        all_comments.extend(next_round_of_comments)
        offset += limit
        # Max offset is 800. The API will return fewer than 'limit' comments of there are no more
        all_comments_scraped = len(next_round_of_comments) < limit or offset > 800
    return all_comments


if __name__ == "__main__":
   if len(sys.argv) <= 1:
      print "Usage: scrape_song.py <url>"
      exit()
   parser = SongDataParser()
   parser.scrape_song(sys.argv[1])


